/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::KITProstheticHandUnit
 * @author     Raphael Grimm ( raphael dot grimm at kit dot edu )
 * @date       2018
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include <thread>
#include <regex>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include "KITProstheticHandUnit.h"

using namespace armarx;


armarx::PropertyDefinitionsPtr KITProstheticHandUnit::createPropertyDefinitions()
{
    return armarx::PropertyDefinitionsPtr(new KITProstheticHandUnitPropertyDefinitions(
            getConfigIdentifier()));
}

void KITProstheticHandUnit::onInitHandUnit()
{
    _driver = std::make_unique<BLEProthesisInterface>(getProperty<std::string>("MAC"));
    //addShapeName("Open"); //is added by something else already
    addShapeName("Close");
    addShapeName("G0");
    addShapeName("G1");
    addShapeName("G2");
    addShapeName("G3");
    addShapeName("G4");
    addShapeName("G5");
    addShapeName("G6");
    addShapeName("G7");
    addShapeName("G8");
}

void KITProstheticHandUnit::onExitHandUnit()
{
    _driver.reset();
}

void KITProstheticHandUnit::setJointAngles(const NameValueMap& targetJointAngles, const Ice::Current&)
{
    ARMARX_CHECK_NOT_NULL(_driver);

    for (const std::pair<std::string, float>& pair : targetJointAngles)
    {
        if (pair.first == "Fingers")
        {
            _driver->sendFingerPWM(200, 2999, static_cast<std::uint64_t>(pair.second * _driver->maxPosF()));
            // fix until hw driver is fixed to handle multiple commands at the same time
            std::this_thread::sleep_for(std::chrono::milliseconds(100));
        }
        else if (pair.first == "Thumb")
        {
            _driver->sendThumbPWM(200, 2999, static_cast<std::uint64_t>(pair.second * _driver->getMaxPosThumb()));
            // fix until hw driver is fixed to handle multiple commands at the same time
            std::this_thread::sleep_for(std::chrono::milliseconds(100));
        }
        else
        {
            ARMARX_WARNING << "Invalid HandJointName '" << pair.first << "', ignoring.";
        }
    }
}

NameValueMap KITProstheticHandUnit::getCurrentJointValues(const Ice::Current&)
{
    NameValueMap jointValues;
    jointValues["Fingers"] = _driver->getFingerPos() * 1.f / _driver->maxPosF();
    jointValues["Thumb"] = _driver->getThumbPos() * 1.f / _driver->getMaxPosThumb();
    return jointValues;
}

void KITProstheticHandUnit::addShape(const std::string& name, const std::map<std::string, float>& shape)
{
    _shapes[name] = shape;
    addShapeName(name);
}

void KITProstheticHandUnit::addShapeName(const std::string& name)
{
    Variant currentPreshape;
    currentPreshape.setString(name);
    shapeNames->addVariant(currentPreshape);
}

void KITProstheticHandUnit::setShape(const std::string& shapeName, const Ice::Current& c)
{
    if (std::regex_match(shapeName, std::regex{"[gG](0|[1-9][0-9]*)"}))
    {
        _driver->sendGrasp(std::stoul(shapeName.substr(1)));
    }
    else if (shapeName == "Open")
    {
        _driver->sendGrasp(0);
    }
    else if (shapeName == "Close")
    {
        _driver->sendGrasp(1);
    }
    else if (!_shapes.count(shapeName))
    {
        ARMARX_WARNING << "Unknown shape name '" << shapeName
                       << "'\nKnown shapes: " << _shapes;
    }
    else
    {
        setJointAngles(_shapes.at(shapeName));
    }
}

